/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggVorbis 'TREMOR' CODEC SOURCE CODE.   *
 *                                                                  *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE OggVorbis 'TREMOR' SOURCE CODE IS (C) COPYRIGHT 1994-2002    *
 * BY THE Xiph.Org FOUNDATION http://www.xiph.org/                  *
 *                                                                  *
 ********************************************************************

 function: normalized modified discrete cosine transform
           power of two length transform only [64 <= n ]
 last mod: $Id: mdct.c,v 1.2 2002/09/03 03:15:19 xiphmont Exp $

 Original algorithm adapted long ago from _The use of multirate filter
 banks for coding of high quality digital audio_, by T. Sporer,
 K. Brandenburg and B. Edler, collection of the European Signal
 Processing Conference (EUSIPCO), Amsterdam, June 1992, Vol.1, pp
 211-214

 The below code implements an algorithm that no longer looks much like
 that presented in the paper, but the basic structure remains if you
 dig deep enough to see it.

 This module DOES NOT INCLUDE code to generate/apply the window
 function.  Everybody has their own weird favorite including me... I
 happen to like the properties of y=sin(2PI*sin^2(x)), but others may
 vehemently disagree.

 ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ivorbiscodec.h"
#include "os.h"
#include "mdct.h"
#include "mdct_lookup.h"
#include "misc.h"

//#define C_VERSION

#ifdef C_VERSION

typedef struct {
  int n;
  
  DATA_TYPE   *trig;
  ogg_int32_t *bitrev;

  /*DATA_TYPE scale;*/
} mdct_lookup;


static void mdct_init(mdct_lookup *lookup,int n){
  lookup->n=n;
  switch(n){
  case 64:
    lookup->trig=triglook_64;
    lookup->bitrev=bitrevlook_64;
    /*lookup->scale=0x04000000;*/
    break;
  case 128:
    lookup->trig=triglook_128;
    lookup->bitrev=bitrevlook_128;
    /*lookup->scale=0x02000000;*/
    break;
  case 256:
    lookup->trig=triglook_256;
    lookup->bitrev=bitrevlook_256;
    /*lookup->scale=0x01000000;*/
    break;
  case 512:
    lookup->trig=triglook_512;
    lookup->bitrev=bitrevlook_512;
    /*lookup->scale=0x00800000;*/
    break;
  case 1024:
    lookup->trig=triglook_1024;
    lookup->bitrev=bitrevlook_1024;
    /*lookup->scale=0x00400000;*/
    break;
  case 2048:
    lookup->trig=triglook_2048;
    lookup->bitrev=bitrevlook_2048;
    /*lookup->scale=0x00200000;*/
    break;
  case 4096:
    lookup->trig=triglook_4096;
    lookup->bitrev=bitrevlook_4096;
    /*lookup->scale=0x00100000;*/
    break;
  case 8192:
    lookup->trig=triglook_8192;
    lookup->bitrev=bitrevlook_8192;
    /*lookup->scale=0x00080000;*/
    break;
  default:
    /* die horribly */
    memset(lookup,0,sizeof(*lookup));
  }
}

/* 8 point butterfly (in place, 4 register) */
STIN void mdct_butterfly_8(DATA_TYPE *x){
  REG_TYPE a0   = x[6] + x[2];
  REG_TYPE a1   = x[6] - x[2];
  REG_TYPE a2   = x[4] + x[0];
  REG_TYPE a3   = x[4] - x[0];

	   x[6] = a0   + a2;
	   x[4] = a0   - a2;
	   
	   a0   = x[5] - x[1];
	   a2   = x[7] - x[3];
	   x[0] = a1   + a0;
	   x[2] = a1   - a0;
	   
	   a0   = x[5] + x[1];
	   a1   = x[7] + x[3];
	   x[3] = a2   + a3;
	   x[1] = a2   - a3;
	   x[7] = a1   + a0;
	   x[5] = a1   - a0;
}

/* 16 point butterfly (in place, 4 register) */
STIN void mdct_butterfly_16(DATA_TYPE *x){
  REG_TYPE a0;
  REG_TYPE a1;
  DATA_TYPE *x2 = x + 8;

           x  += 6;
           x2 += 6;
/*0*/
           a0     = x2[0] - x[0];
           a1     = x2[1] - x[1];
           x2[0] += x[0];
           x2[1] += x[1];
           x[0]   = a0;
           x[1]   = a1;
           x  -= 2;
           x2 -= 2;
/*1*/
           a0     = x2[0] - x[0];
           a1     = x2[1] - x[1];
           x2[0] += x[0];
           x2[1] += x[1];
           x[0]   = MULT31((a0   - a1) , cPI2_8);
           x[1]   = MULT31((a0   + a1) , cPI2_8);
           x  -= 2;
           x2 -= 2;
/*2*/
           a0     = x[1]  - x2[1];
           a1     = x2[0] - x[0];
           x2[0] += x[0];
           x2[1] += x[1];
           x[0]   = a0;
           x[1]   = a1;
           x  -= 2;
           x2 -= 2;
/*3*/
           a0     = x[1]  - x2[1];
           a1     = x[0]  - x2[0];
           x2[0] += x[0];
           x2[1] += x[1];
           x[0]   = MULT31((a0   + a1) , cPI2_8);
           x[1]   = MULT31((a0   - a1) , cPI2_8);

	   mdct_butterfly_8(x);
	   mdct_butterfly_8(x+8);
}

/* 32 point butterfly (in place, 4 register) */
STIN void mdct_butterfly_32(DATA_TYPE *x){
  REG_TYPE a0;
  REG_TYPE a1;
  DATA_TYPE *x2 = x;

/*0*/
           x  += 14;
           x2 += 30;
           a0     = x2[0] - x[0];
           a1     = x2[1] - x[1];
           x2[0] +=         x[0];           
	   x2[1] +=         x[1];
           x[0]   =         a0;              
	   x[1]   =         a1;
	   x  -= 2;
	   x2 -= 2;
/*1*/
           a0     = x2[0] - x[0];
	   a1     = x2[1] - x[1];
           x2[0] +=         x[0];
	   x2[1] +=         x[1];
           x[0]   = (MULT32( a0 , cPI1_8 ) - MULT32( a1 , cPI3_8 ))<<1;
	   x[1]   = (MULT32( a0 , cPI3_8 ) + MULT32( a1 , cPI1_8 ))<<1;
	   x  -= 2;
	   x2 -= 2;
/*2*/
           a0     = x2[0] - x[0];
	   a1     = x2[1] - x[1];
	   x2[0] +=         x[0];
	   x2[1] +=         x[1];
	   x[0]   = MULT31(( a0  - a1 ) , cPI2_8);
	   x[1]   = MULT31(( a0  + a1 ) , cPI2_8);
	   x  -= 2;
	   x2 -= 2;
/*3*/
	   a0     = x2[0] - x[0];
	   a1     = x2[1] - x[1];
	   x2[0] += x[0];
	   x2[1] += x[1];
	   x[0]   = (MULT32( a0 , cPI3_8 ) - MULT32( a1 , cPI1_8 ))<<1;
	   x[1]   = (MULT32( a1 , cPI3_8 ) + MULT32( a0 , cPI1_8 ))<<1;
	   x  -= 2;
	   x2 -= 2;
/*4*/
	   a0     = x2[0] - x[0];
	   a1     = x[1]  - x2[1];
	   x2[0] += x[0];
	   x2[1] += x[1];
	   x[0]   = a1;
	   x[1]   = a0;
	   x  -= 2;
	   x2 -= 2;
/*5*/
	   a0     = x[0]  - x2[0];
	   a1     = x[1]  - x2[1];
	   x2[0] += x[0];
	   x2[1] += x[1];
	   x[0]   = (MULT32( a1 , cPI1_8 ) + MULT32( a0 , cPI3_8 ))<<1;
	   x[1]   = (MULT32( a1 , cPI3_8 ) - MULT32( a0 , cPI1_8 ))<<1;
	   x  -= 2;
	   x2 -= 2;
/*6*/
	   a0     = x[0]  - x2[0];
	   a1     = x[1]  - x2[1];
	   x2[0] += x[0];
	   x2[1] += x[1];
	   x[0]   = MULT31(( a1  + a0 ) , cPI2_8);
	   x[1]   = MULT31(( a1  - a0 ) , cPI2_8);
	   x  -= 2;
	   x2 -= 2;
/*7*/
	   a0     = x[0]  - x2[0];
	   a1     = x[1]  - x2[1];
	   x2[0] += x[0];
	   x2[1] += x[1];
	   x[0]   = (MULT32( a1 , cPI3_8 ) + MULT32( a0 , cPI1_8 ))<<1;
	   x[1]   = (MULT32( a1 , cPI1_8 ) - MULT32( a0 , cPI3_8 ))<<1;

	   mdct_butterfly_16(x);
	   mdct_butterfly_16(x+16);

}

/* N/stage point generic N stage butterfly (in place, 2 register) */
STIN void mdct_butterfly_generic(int points,
                                 int trigint,
                                 DATA_TYPE *x,
                                 DATA_TYPE *T){
  
  DATA_TYPE *x1        = x          + points;
  DATA_TYPE *x2        = x          + (points>>1);
  REG_TYPE   a0;
  REG_TYPE   a1;

  do{
    x1 -= 2;
    x2 -= 2;
               a0      = x1[0]      -  x2[0];
	       a1      = x1[1]      -  x2[1];
	       x1[0]  += x2[0];
	       x1[1]  += x2[1];
	       x2[0]   = (MULT32(a1 , T[1] ) + MULT32(a0 , T[0]))<<2;
	       x2[1]   = (MULT32(a1 , T[0] ) - MULT32(a0 , T[1]))<<2;
	       
	       T+=trigint<<2;
    x1 -= 2;
    x2 -= 2;
	       
	       a0      = x1[0]      -  x2[0];
	       a1      = x1[1]      -  x2[1];
	       x1[0]  += x2[0];
	       x1[1]  += x2[1];
	       x2[0]   = (MULT32(a1 , T[1] ) + MULT32( a0 , T[0]))<<2;
	       x2[1]   = (MULT32(a1 , T[0] ) - MULT32( a0 , T[1]))<<2;
	       
	       T+=trigint<<2;
    x1 -= 2;
    x2 -= 2;
	       
	       a0      = x1[0]      -  x2[0];
	       a1      = x1[1]      -  x2[1];
	       x1[0]  += x2[0];
	       x1[1]  += x2[1];
	       x2[0]   = (MULT32(a1 , T[1] ) + MULT32( a0 , T[0]))<<2;
	       x2[1]   = (MULT32(a1 , T[0] ) - MULT32( a0 , T[1]))<<2;
	       
	       T+=trigint<<2;
    x1 -= 2;
    x2 -= 2;
	       
	       a0      = x1[0]      -  x2[0];
	       a1      = x1[1]      -  x2[1];
	       x1[0]  += x2[0];
	       x1[1]  += x2[1];
	       x2[0]   = (MULT32(a1 , T[1] ) + MULT32( a0 , T[0]))<<2;
	       x2[1]   = (MULT32(a1 , T[0] ) - MULT32( a0 , T[1]))<<2;

	       T+=trigint<<2;
  }while(x2>x);
}

STIN void mdct_butterflies(mdct_lookup *init,
			   DATA_TYPE *x,
			   int points)
{
  int j;
  DATA_TYPE *T=init->trig;
  int i;
  int p = points;
  i = 1;
  if (p != 32)
  {
      do
      {
          DATA_TYPE *xx = x;
          
          for (j = i; j > 0; j--)
          {
              mdct_butterfly_generic(p,i,xx,T);
              xx += p;
          }
          p>>=1;
          i<<=1;
      }
      while (p != 32);
  }
  for(j=points;j> 0;j-=32)
  {
    mdct_butterfly_32(x);
    x += 32;
  }
}

STIN void mdct_bitreverse(mdct_lookup *init, 
			    DATA_TYPE *x){
  int          n     = init->n;
  ogg_int32_t *bit   = init->bitrev;
  DATA_TYPE   *w0    = x;
  DATA_TYPE   *w1    = x = w0+(n>>1);
  DATA_TYPE   *T     = init->trig+n;

  do{
    DATA_TYPE *x0    = x+bit[0];
    DATA_TYPE *x1    = x+bit[1];

    REG_TYPE  a0     = x0[1]  - x1[1];
    REG_TYPE  a1     = x0[0]  + x1[0];
    REG_TYPE  a2     = (MULT32(a1     , T[0] )  + MULT32(a0 , T[1]))<<2;
    REG_TYPE  a3     = (MULT32(a1     , T[1] )  - MULT32(a0 , T[0]))<<2;

	      w1    -= 4;

              a0     = (x0[1] + x1[1])/2;
              a1     = (x0[0] - x1[0])/2;
      
	      w0[0]  = a0     + a2;
	      w1[2]  = a0     - a2;
	      w0[1]  = a1     + a3;
	      w1[3]  = a3     - a1;

              x0     = x+bit[2];
              x1     = x+bit[3];

              a0     = x0[1]  - x1[1];
              a1     = x0[0]  + x1[0];
              a2     = (MULT32(a1     , T[2] )  + MULT32(a0 , T[3]))<<2;
              a3     = (MULT32(a1     , T[3] )  - MULT32(a0 , T[2]))<<2;

              a0     = (x0[1] + x1[1])/2;
              a1     = (x0[0] - x1[0])/2;
      
	      w0[2]  = a0     + a2;
	      w1[0]  = a0     - a2;
	      w0[3]  = a1     + a3;
	      w1[1]  = a3     - a1;

	      T     += 4;
	      bit   += 4;
	      w0    += 4;

  }while(w0<w1);
}

void mdct_backward(int n, DATA_TYPE *in, DATA_TYPE *out){
  mdct_lookup init;
  DATA_TYPE *iX;
  DATA_TYPE *oX;
  DATA_TYPE *T;

  mdct_init(&init,n);

  /* rotate */

             iX = in+(n>>1)-7;
	     oX = out+(n>>1)+(n>>2);
             T  = init.trig+(n>>2);

  do{
    oX         -= 4;
    oX[0]       = (MULT32(-iX[2] , T[3]) - MULT32( iX[0]  , T[2]))<<2;
    oX[1]       = (MULT32( iX[0] , T[3]) + MULT32(-iX[2]  , T[2]))<<2;
    oX[2]       = (MULT32(-iX[6] , T[1]) - MULT32( iX[4]  , T[0]))<<2;
    oX[3]       = (MULT32( iX[4] , T[1]) + MULT32(-iX[6]  , T[0]))<<2;
    iX         -= 8;
    T          += 4;
  }while(iX>=in);

  iX            = in+(n>>1)-8;
  oX            = out+(n>>1)+(n>>2);
  T             = init.trig+(n>>2);

  do{
    T          -= 4;
    oX[0]       =  (MULT32 (iX[4] , T[3]) + MULT32(iX[6] , T[2]))<<2;
    oX[1]       =  (MULT32 (iX[4] , T[2]) - MULT32(iX[6] , T[3]))<<2;
    oX[2]       =  (MULT32 (iX[0] , T[1]) + MULT32(iX[2] , T[0]))<<2;
    oX[3]       =  (MULT32 (iX[0] , T[0]) - MULT32(iX[2] , T[1]))<<2;
    iX         -= 8;
    oX         += 4;
  }while(iX>=in);

  mdct_butterflies(&init,out+(n>>1),(n>>1));
  mdct_bitreverse(&init,out);

  /* roatate + window */

  {
    DATA_TYPE *oX1;
    DATA_TYPE *oX2;
    DATA_TYPE *iX;

    oX1=out+(n>>1)+(n>>2);    
    oX2=out+(n>>1)+(n>>2);
    iX =out;
    T  =init.trig+(n>>1);
    do{
      oX1-=4;
      oX1[3]  =  (MULT32 (iX[0] , T[1]) - MULT32(iX[1] , T[0]))<<2;
      oX2[0]  =-((MULT32 (iX[0] , T[0]) + MULT32(iX[1] , T[1]))<<2);

      oX1[2]  =  (MULT32 (iX[2] , T[3]) - MULT32(iX[3] , T[2]))<<2;
      oX2[1]  =-((MULT32 (iX[2] , T[2]) + MULT32(iX[3] , T[3]))<<2);

      oX1[1]  =  (MULT32 (iX[4] , T[5]) - MULT32(iX[5] , T[4]))<<2;
      oX2[2]  =-((MULT32 (iX[4] , T[4]) + MULT32(iX[5] , T[5]))<<2);

      oX1[0]  =  (MULT32 (iX[6] , T[7]) - MULT32(iX[7] , T[6]))<<2;
      oX2[3]  =-((MULT32 (iX[6] , T[6]) + MULT32(iX[7] , T[7]))<<2);
      oX2+=4;
      iX    +=   8;
      T     +=   8;
    }while(iX<oX1);

    iX=out+(n>>1)+(n>>2);
    oX1=out+(n>>2);
    oX2=oX1;

    do{
      oX1-=4;
      iX-=4;

      oX2[0] = -(oX1[3] = iX[3]);
      oX2[1] = -(oX1[2] = iX[2]);
      oX2[2] = -(oX1[1] = iX[1]);
      oX2[3] = -(oX1[0] = iX[0]);

      oX2+=4;
    }while(oX2<iX);

    iX=out+(n>>1)+(n>>2);
    oX1=out+(n>>1)+(n>>2);
    oX2=out+(n>>1);
    do{
      oX1-=4;
      oX1[0]= iX[3];
      oX1[1]= iX[2];
      oX1[2]= iX[1];
      oX1[3]= iX[0];
      iX+=4;
    }while(oX1>oX2);
  }
}
#endif
